# -*- coding: utf-8 -*-

import re

BLOCK_START_STRING = '{%'
BLOCK_END_STRING = '%}'
VARIABLE_START_STRING = '{{'
VARIABLE_END_STRING = '}}'
TAGTEST = '%s.+%s|%s.+%s' % (BLOCK_START_STRING, BLOCK_END_STRING, VARIABLE_START_STRING, VARIABLE_END_STRING)
XVTEST = '^ *%s *xv.+%s *$' % (BLOCK_START_STRING, BLOCK_END_STRING)
VTEST = '^%s.+%s$' % (VARIABLE_START_STRING, VARIABLE_END_STRING)
BLOCKTEST = '%s.+%s' % (BLOCK_START_STRING, BLOCK_END_STRING)

CELL = 'cell'
BEFOREROW = 'beforerow'
BEFORECELL = 'beforecell'
AFTERCELL = 'aftercell'
RANGE = 'range'
BEFORERANGE = 'beforerange'
AFTERRANGE = 'afterrange'
COORD = '[a-z]{1,3}\d+'
RANGETAG = '%s *%s *(%s:%s\S*?) *%s' % (RANGE, VARIABLE_START_STRING, COORD, COORD, VARIABLE_END_STRING)
CELLTAG = '%s *%s *(%s) *%s' % (CELL, VARIABLE_START_STRING, COORD, VARIABLE_END_STRING)
CELLSEPERATORS = [BEFOREROW, BEFORECELL, AFTERCELL]
CELLSPLITTER = '(%s)' % '|'.join(CELLSEPERATORS)
RANGESEPERATORS = [BEFORERANGE, AFTERRANGE]
RANGESPLITTER = '(%s)' % '|'.join(RANGESEPERATORS)

def tag_test(txt):
    p = re.compile(TAGTEST)
    rv = p.findall(txt)
    return bool(rv)

def xv_test(txt):
    p = re.compile(XVTEST)
    xv = p.findall(txt)
    return bool(xv)

def v_test(txt):
    p = re.compile(VTEST)
    v = p.findall(txt)
    return bool(v)

def block_tag_test(txt):
    p = re.compile(BLOCKTEST)
    rv = p.findall(txt)
    return bool(rv)

def split(splitter, txt):
    parts = splitter.split(txt)
    d = {}
    for i in range(1, len(parts), 2):
            d[parts[i]] = parts[i + 1].strip()
    return d

def parse_cell_tag(txt):
    cell_pattern = re.compile(CELLTAG, re.I)
    split_pattern = re.compile(CELLSPLITTER, re.I)
    parts = cell_pattern.split(txt)
    if len(parts) > 1:
        tag = parts[2]
        tag_map = split(split_pattern, tag)
        return parts[1], tag_map
    else:
        tag = txt
        tag_map = split(split_pattern, tag)
        return None, tag_map
    return None, None

def parse_range_tag(txt):
    range_pattern = re.compile(RANGETAG, re.I)
    split_pattern = re.compile(RANGESPLITTER, re.I)
    parts = range_pattern.split(txt)
    if len(parts) > 1:
        tag = parts[2]
        tag_map = split(split_pattern, tag)
        return parts[1], tag_map
    return None, None


CTEx = ' *yn | *xv | *img '
CTBeforeRow = '^(%s-(?!%s).+?%s)+' % (BLOCK_START_STRING, CTEx, BLOCK_END_STRING)
CTBeforeCell = '^(%s(?!%s).+?%s)+' % (BLOCK_START_STRING, CTEx, BLOCK_END_STRING)
CTAfterCell = '(%s(?!%s).+?%s)+$' % (BLOCK_START_STRING, CTEx, BLOCK_END_STRING)
CTExtraCell = '(%s\+(?!%s).+?%s)+$' % (BLOCK_START_STRING, CTEx, BLOCK_END_STRING)
p_br = re.compile(CTBeforeRow)
p_bc = re.compile(CTBeforeCell)
p_ac = re.compile(CTAfterCell)
p_ex = re.compile(CTExtraCell)

from .misc import CellTag
def find_tag(pattern, string):
    m = pattern.search(string)
    if m:
        st, end = m.span()
        tag = string[st:end]
        if st == 0:
            string = string[end:]
        else:
            string = string[:st]
        return string, tag, end-st
    return string, '', 0

def find_cell_tag(s):
    head = 0
    tail = 0
    cell_tag = CellTag()
    s, tag, l = find_tag(p_br, s)
    cell_tag.beforerow = tag
    head += l
    s, tag, l = find_tag(p_bc, s)
    cell_tag.beforecell = tag
    head += l
    s, tag, l = find_tag(p_ex, s)
    cell_tag.extracell = tag
    tail += l
    s, tag, l = find_tag(p_ac, s)
    cell_tag.aftercell = tag
    tail += l
    if head > 0 or tail >0:
        return s, cell_tag, head, tail
    else:
        return s, None, 0, 0

BLOCKSPLIT = '((?:%s(?:(?!%s).)+?%s)+)' % (BLOCK_START_STRING, CTEx, BLOCK_END_STRING)
CTEx2 = ' *yn | *img '
YNSPLIT = '(%s(?:(?=%s)).+?%s)' % (BLOCK_START_STRING, CTEx2, BLOCK_END_STRING)
IMGTEST = '%s *img .+%s' % (BLOCK_START_STRING, BLOCK_END_STRING)

def block_split(txt):
    split_pattern = re.compile(BLOCKSPLIT)
    parts = split_pattern.split(txt)
    return parts

def rich_split(txt):
    split_pattern = re.compile(YNSPLIT)
    parts = split_pattern.split(txt)
    return parts

def img_test(txt):
    p = re.compile(IMGTEST)
    rv = p.findall(txt)
    return bool(rv)


FIXTEST = '({(?:___\d+___)?(?:{|%).+?(?:}|%)(?:___\d+___)?})'
RUNSPLIT = '(___\d+___)'
RUNSPLIT2 = '___(\d+)___'

#need a better way to handle this
def fix_test(txt):
    split_pattern = re.compile(FIXTEST)
    parts = split_pattern.split(txt)
    for i, part in enumerate(parts):
        if i % 2 == 1:
            pattern = re.compile(RUNSPLIT)
            rv = pattern.findall(part)
            if rv :
                return True

def tag_fix(txt):
    split_pattern = re.compile(FIXTEST)
    parts = split_pattern.split(txt)
    p = ''
    for i, part in enumerate(parts):
        if i % 2 == 1:
            p += fix_step2(part)
        else:
            p += part
    split_pattern2 = re.compile(RUNSPLIT2)
    parts = split_pattern2.split(p)
    d = {}
    for i in range(1, len(parts), 2):
            d[int(parts[i])] = parts[i + 1]
    return d

def fix_step2(txt):
    split_pattern = re.compile(RUNSPLIT)
    parts = split_pattern.split(txt)
    p0 = ''
    p1 = ''
    for index,part in enumerate(parts):
        if index % 2 == 0:
            p0 += part
        else:
            p1 += part
    return p0+p1



# -*- coding: utf-8 -*-

import six
from .utils import tag_test, xv_test, v_test, find_cell_tag, block_split, rich_split, img_test
from .nodemap import node_map

class Node(object):
    ext_tag = 'node'

    def __init__(self):
        self._children = []
        self._parent = None
        self.no = 0
        self._depth = -1

    @property
    def depth(self):
        if self._depth == -1:
            if self._parent is None or self._parent is self:
                self._depth = 0
            else:
                self._depth = self._parent.depth + 1
        return self._depth

    @property
    def node_key(self):
        return id(self)
        return '%s,%d' % (self._parent.node_key, self.no)

    @property
    def node_tag(self):
        return "{%%%s %s %%}" % (self.ext_tag, self.node_key)
        return "{%%%s '%s' %%}" % (self.ext_tag, self.node_key)

    @property
    def print_tag(self):
        return self.node_tag

    def children_to_tag(self):
        x = []
        for child in self._children:
            x.append(child.to_tag())
        return '\n'.join(x)

    def to_tag(self):
        if self._children:
            return self.children_to_tag()
        node_map.put(self.node_key, self)
        return self.node_tag

    def tag_tree(self):
        print('\t' * self.depth, self.print_tag)
        for child in self._children:
            child.tag_tree()

    def add_child(self, child):
        child.no = len(self._children)
        child._parent = self
        self._children.append(child)

    def enter(self):
        pass

    def reenter(self):
        self.enter()

    def child_reenter(self):
        pass

    def exit(self):
        pass

    def __str__(self):
        return self.__class__.__name__ + ' , ' + self.node_tag

    def set_image_ref(self, image_ref, image_key):
        self._parent.set_image_ref(image_ref, image_key)

class Segment(Node):

    def __init__(self, text):
        Node.__init__(self)
        self.text = text

    @property
    def node_tag(self):
        #fmt = "{%%seg '%s'%%}%s{%%endseg%%}"
        fmt = "{%%seg %d%%}%s{%%endseg%%}"
        return fmt % (self.node_key, self.text)

    def process_rv(self, rv):
        self._parent.process_child_rv(rv)
        return rv

class RichSegment(Segment):

    def __init__(self, text, font):
        Segment.__init__(self, text)
        self.font = font

    def process_rv(self, rv):
        self._parent.process_child_rv(self.rv)
        return self.text

    def process_rich_rv(self, rv):
        self.rv = rv
        return self.text

class BlockSegment(Segment):

    @property
    def node_tag(self):
        return self.text

class ImageSegment(Segment):

    @property
    def node_tag(self):
        #fmt = "{%%seg '%s'%%}{%%endseg%%}%s"
        fmt = "{%%seg %s%%}{%%endseg%%}%s"
        return fmt % (self.node_key, self.text)

class Section(Node):

    def __init__(self, text, font, rich_handler):
        Node.__init__(self)
        self.font = font
        self.rich_handler = rich_handler
        self.unpack(text)

    def unpack(self, text):
        parts = block_split(text)
        for index, part in enumerate(parts):
            if part == '':
                continue
            if index % 2 == 0:
                sub_parts = rich_split(part)
                for sub_index, sub_part in enumerate(sub_parts):
                    if sub_part == '':
                        continue
                    if sub_index % 2 == 0:
                        child = Segment(sub_part)
                    else:
                        if img_test(sub_part):
                            child = ImageSegment(sub_part)
                        else:
                            child = RichSegment(sub_part, self.font)
                    self.add_child(child)
            else:
                child = BlockSegment(part)
                self.add_child(child)

    def pack(self):
        if not self.richs:
            text = ''.join(self.child_rvs)
            if isinstance(self, TagCell):
                return text
            else:
                return self.rich_handler.rich_segment(text, self.font)
        richs = self.richs
        rvs = self.child_rvs
        font = self.font
        richs.append(len(rvs))
        st = 0
        rv = []
        for i in richs:
            if st == i:
                pass
            else:
                if st + 1 == i:
                    text = rvs[st]
                else:
                    slice = rvs[st:i]
                    text = ''.join(slice)
                rich_text = self.rich_handler.rich_segment(text, self.font)
                rv.append(rich_text)
            if i < len(rvs):
                rv.append(rvs[i])
            st = i + 1
        return rv

    def enter(self):
        self.child_rvs = []
        self.richs = []

    def exit(self):
        rv = self.pack()
        self._parent.process_child_rv(rv)

    def process_child_rv(self, rv):
        if not isinstance(rv, six.text_type):
            self.richs.append(len(self.child_rvs))
        self.child_rvs.append(rv)

class Cell(Node):
    ext_tag = 'cell'

    def __init__(self, sheet_cell, rowx, colx, value, cty):
        Node.__init__(self)
        self.sheet_cell = sheet_cell
        self.rowx = rowx
        self.colx = colx
        self.value = value
        self.cty = cty
        self.cell_tag = None

    def to_tag(self):
        tag = Node.to_tag(self)
        if self.cell_tag:
            beforecell = self.cell_tag.beforecell
            aftercell = self.cell_tag.aftercell
            return beforecell + tag + aftercell
        return tag

    def exit(self):
        self.write(self.value, self.cty)

    def process_child_rv(self, rv):
        pass

    def write(self, rv, cty):
        self._parent.write_cell(self, rv, cty)

    def set_image_ref(self, image_ref, image_key):
        self._parent.set_image_ref(image_ref, (self.rowx,self.colx,image_key))

class TagCell(Section, Cell):

    def __init__(self, sheet_cell, rowx, colx, value, cty, font, rich_handler):
        Cell.__init__(self, sheet_cell, rowx, colx, value, cty)
        self.font = font
        self.rich_handler = rich_handler
        self.unpack(value)

    def exit(self):
        rv = self.pack()
        if not isinstance(rv, six.text_type):
            rv = self.rich_handler.rich_content(rv)
        self.write(rv, self.cty)


class RichTagCell(Cell):

    def __init__(self, sheet_cell, rowx, colx, value, cty, font, rich_handler):
        Cell.__init__(self, sheet_cell, rowx, colx, value, cty)
        self.font = font
        self.rich_handler = rich_handler
        self.unpack(value)

    def unpack(self, rich_text):
        for text, font, segment in self.rich_handler.iter(rich_text, self.font):
            section = Section(text, font, self.rich_handler)
            self.add_child(section)

    def enter(self):
        self.child_rvs = []

    def exit(self):
        rv = self.rich_handler.rich_content(self.child_rvs)
        self.write(rv, self.cty)

    def process_child_rv(self, rv):
        if isinstance(rv, list):
            self.child_rvs.extend(rv)
        else:
            self.child_rvs.append(rv)

class EmptyCell(Cell):

    def __init__(self, rowx, colx):
        Node.__init__(self)
        self.sheet_cell = None
        self.rowx = rowx
        self.colx = colx
        self.value = None
        self.cty = None
        self.cell_tag = None

class XvCell(Cell):

    def __init__(self, sheet_cell, rowx, colx, value, cty, isXv):
        Cell.__init__(self, sheet_cell, rowx, colx, value, cty)
        self.isXv = isXv

    @property
    def node_tag(self):
        tag = self.value.strip()
        if self.isXv:
            head = tag[:-2].strip()
            tag = "%s,%d%%}" % (head, self.node_key)
        else:
            body = tag[2:-2].strip()
            tag = "{%%xv %s,%d%%}" % (body, self.node_key)
        return tag

    def enter(self):
        self.rv = None

    def exit(self):
        self.write(self.rv, None)


class Row(Node):
    ext_tag = 'row'

    def __init__(self, rowx):
        Node.__init__(self)
        self.rowx = rowx
        self.cell_tag = None

    def to_tag(self):
        tag = Node.to_tag(self)
        if self.cell_tag:
            beforerow = self.cell_tag.beforerow
            return beforerow + tag
        return tag

    def enter(self):
        self._parent.write_row(self)

class Tree(Node):
    ext_tag = 'tree'

    def __init__(self, index):
        Node.__init__(self)
        self._depth = 0
        self.no = index
        self.sheet_writer = None

    @property
    def node_key(self):
        return str(self.no)

    def set_sheet_writer(self, sheet_writer):
        self.sheet_writer = sheet_writer

    def write_row(self, row_node):
        self.sheet_writer.write_row(row_node)

    def write_cell(self, cell_node, rv, cty):
        self.sheet_writer.write_cell(cell_node, rv, cty)

    def set_image_ref(self, image_ref, image_key):
        self.sheet_writer.set_image_ref(image_ref, image_key)

def create_cell(sheet_cell, rowx, colx, value, rich_text, data_type, font, rich_handler):
    s,cell_tag,head,tail = find_cell_tag(value)
    if s == '':
        cell = Cell(sheet_cell, rowx, colx, s, data_type)
    elif xv_test(s):
        cell = XvCell(sheet_cell, rowx, colx, s, data_type, True)
    elif v_test(s):
        cell = XvCell(sheet_cell, rowx, colx, s, data_type, False)
    elif not rich_text:
        cell = TagCell(sheet_cell, rowx, colx, s, data_type, font, rich_handler)
    else:
        if head == 0 and tail == 0:
            cell = RichTagCell(sheet_cell, rowx, colx, rich_text, data_type, font, rich_handler)
        else:
            _tail = head + len(s) - 1
            _rich,_text = rich_handler.mid(rich_text, head, _tail)
            cell = RichTagCell(sheet_cell, rowx, colx, _rich, data_type, font, rich_handler)
    if cell_tag:
        cell.cell_tag = cell_tag
    return cell



# -*- coding: utf-8 -*-
import sys
from jinja2 import Environment
from jinja2.exceptions import TemplateSyntaxError

class Env(Environment):

    def handle_exception(self, *args, **kwargs):
        exc_type, exc_value, tb = sys.exc_info()
        red_fmt = '\033[31m%s\033[0m'
        blue_fmt = '\033[34m%s\033[0m'
        error_type = red_fmt % ('error type:  %s' % exc_type)
        error_message = red_fmt % ('error message:  %s' % exc_value)
        print(error_type)
        print(error_message)
        if exc_type is TemplateSyntaxError:
            lineno = exc_value.lineno
            source = kwargs['source']
            src_lines = source.splitlines()
            for i, line in enumerate(src_lines):
                if i + 1 == lineno:
                    line_str = red_fmt % ('line %d : %s' % (i + 1, line))
                elif i + 1 in [lineno - 1, lineno + 1]:
                    line_str = blue_fmt % ('line %d : %s' % (i + 1, line))
                else:
                    line_str = 'line %d : %s' % (i + 1, line)
                print(line_str)
        Environment.handle_exception(self, *args, **kwargs)

class TreeProperty(object):

    def __init__(self, name):
        self.name = name
        self._name = '_' + name

    def __set__(self, instance, value):
        instance.__dict__[self._name] = value

    def __get__(self, instance, cls):
        if not hasattr(instance, self._name):
            instance.__dict__[self._name] = getattr(instance._parent, self.name)
        return instance.__dict__[self._name]

class CellTag():

    def __init__(self, cell_tag=dict()):
        self.beforerow = ''
        self.beforecell = ''
        self.aftercell = ''
        self.extracell = ''
        if cell_tag:
            self.__dict__.update(cell_tag)

    def extend(self, other):
        if isinstance(other, CellTag):
            self.beforerow = other.beforerow + self.beforerow
            self.beforecell = other.beforecell + self.beforecell
            self.aftercell += other.aftercell
            self.extracell += other.extracell



# -*- coding: utf-8 -*-

import six

from .base import SheetBase, BookBase
from .writermixin import SheetMixin, BookMixin
from .utils import tag_test
from .xlnode import Tree, Row, Cell, EmptyCell, Node, create_cell
from .jinja import jinja_env
from .sheetresource import SheetResource
from .richtexthandler import rich_handler
from .merger import Merger

class SheetWriter(SheetBase, SheetMixin):

    def __init__(self, bookwriter, sheet_resource, sheet_name):
        self.rdbook = bookwriter.rdbook
        self.wtbook = bookwriter.wtbook
        self.style_list = bookwriter.style_list
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet
        self.create_worksheet(self.rdsheet, sheet_name)
        self.wtrows = set()
        self.wtcols = set()
        self.min_rowx = -1
        self.min_colx = -1
        self.reset_pos()

class BookWriter(BookBase, BookMixin):
    sheet_writer_cls = SheetWriter

    def __init__(self, fname, debug=False):
        self.debug = debug
        self.load(fname)

    def load(self, fname):
        self.workbook = self.load_rdbook(fname)
        self.font_map = {}
        self.sheet_writer_map = {}
        self.sheet_resource_map = {}
        self.jinja_env = jinja_env
        for index,rdsheet in enumerate(self.rdbook.sheets()):
            sheet_tree = self.build(rdsheet, index)
            merger = Merger(rdsheet)
            sheet_resource = SheetResource(rdsheet, sheet_tree, jinja_env, merger)
            self.put_sheet_resource(index, rdsheet.name, sheet_resource)

    def build(self, sheet, index):
        tree = Tree(index)
        for rowx in range(sheet.nrows):
            row_node = Row(rowx)
            tree.add_child(row_node)
            for colx in range(sheet.ncols):
                try:
                    sheet_cell = sheet.cell(rowx, colx)
                except:
                    cell_node = EmptyCell(rowx, colx)
                    tree.add_child(cell_node)
                    continue
                value = sheet_cell.value
                cty = sheet_cell.ctype
                rich_text = self.get_rich_text(sheet, rowx, colx)
                if isinstance(value, six.text_type):
                    if not tag_test(value):
                        if rich_text:
                            cell_node = Cell(sheet_cell, rowx, colx, rich_text, cty)
                        else:
                            cell_node = Cell(sheet_cell, rowx, colx, value, cty)
                    else:
                        font = self.get_font(sheet, rowx, colx)
                        cell_node = create_cell(sheet_cell, rowx, colx, value, rich_text, cty, font, rich_handler)
                else:
                    cell_node = Cell(sheet_cell, rowx, colx, value, cty)
                tree.add_child(cell_node)
            tree.add_child(Node())  #
        return tree

    def render_sheet(self, payload):
        if not hasattr(self, 'wtbook') or self.wtbook is None:
            self.create_workbook()
        BookMixin.render_sheet(self, payload)

    def save(self, fname):
        if self.wtbook is not None:
            stream = open(fname, 'wb')
            self.wtbook.save(stream)
            stream.close()
            del self.wtbook
        self.sheet_writer_map.clear()




from .nodemap import node_map
class SheetResource():

    def __init__(self, rdsheet, sheet_tree, jinja_env, merger):
        self.rdsheet = rdsheet
        self.sheet_tree = sheet_tree
        self.tpl = sheet_tree.to_tag()
        #print(self.tpl)
        self.merger = merger
        self.jinja_tree = jinja_env.from_string(self.tpl)


    def render_sheet(self, sheet_writer, payload):
        node_map.set_current_node(self.sheet_tree)
        self.sheet_tree.set_sheet_writer(sheet_writer)
        self.jinja_tree.render(payload)
        self.merger.collect_range(sheet_writer.wtsheet)




from .misc import Env
from .xlext import NodeExtension, SegmentExtension, XvExtension, ImageExtension, ImagexExtension
from .ynext import YnExtension, YnxExtension

jinja_env = Env(extensions=[NodeExtension, SegmentExtension, YnExtension,
                            XvExtension, ImageExtension])
jinja_envx = Env(extensions=[NodeExtension, SegmentExtension, YnxExtension,
                            XvExtension, ImagexExtension])



# -*- coding: utf-8 -*-

import xlrd
import xlwt
import six
import datetime
from decimal import Decimal
from openpyxl.utils.datetime import to_excel

from openpyxl.cell.cell import NUMERIC_TYPES, TIME_TYPES, STRING_TYPES
BOOL_TYPE = bool

def get_type(value):
    if isinstance(value, NUMERIC_TYPES):
        dt = xlrd.XL_CELL_NUMBER
    elif isinstance(value, STRING_TYPES):
        dt = xlrd.XL_CELL_TEXT
    elif isinstance(value, TIME_TYPES):
        dt = xlrd.XL_CELL_DATE
        return to_excel(value), dt
    elif isinstance(value, BOOL_TYPE):
        dt = xlrd.XL_CELL_BOOLEAN
    else:
        return str(value), xlrd.XL_CELL_TEXT
    return value, dt

# adapted from xlutils.filter
class SheetBase():

    def create_worksheet(self, rdsheet, wtsheet_name):
        # these checks should really be done by xlwt!
        if not wtsheet_name:
            raise ValueError('Empty sheet name will result in invalid Excel file!')
        l_wtsheet_name = wtsheet_name.lower()
        if l_wtsheet_name in self.wtbook.wtsheet_names:
            raise ValueError('A sheet named %r has already been added!' % l_wtsheet_name)
        self.wtbook.wtsheet_names.add(l_wtsheet_name)
        l_wtsheet_name = len(wtsheet_name)
        if len(wtsheet_name) > 31:
            raise ValueError('Sheet name cannot be more than 31 characters long, '
                             'supplied name was %i characters long!' % l_wtsheet_name)

        self.rdsheet = rdsheet
        self.wtsheet_name = wtsheet_name
        self.wtsheet = wtsheet = self.wtbook.add_sheet(wtsheet_name, cell_overwrite_ok=True)
        wtsheet.mc_ranges = {}

        # default column width: STANDARDWIDTH, DEFCOLWIDTH
        #
        if rdsheet.standardwidth is not None:
            # STANDARDWIDTH is expressed in units of 1/256 of a
            # character-width, but DEFCOLWIDTH is expressed in units of
            # character-width; we lose precision by rounding to
            # the higher whole number of characters.
            #### XXXX TODO: implement STANDARDWIDTH record in xlwt.
            wtsheet.col_default_width = \
                (rdsheet.standardwidth + 255) // 256
        elif rdsheet.defcolwidth is not None:
            wtsheet.col_default_width = rdsheet.defcolwidth
        #
        # WINDOW2
        #
        wtsheet.show_formulas = rdsheet.show_formulas
        wtsheet.show_grid = rdsheet.show_grid_lines
        wtsheet.show_headers = rdsheet.show_sheet_headers
        wtsheet.panes_frozen = rdsheet.panes_are_frozen
        wtsheet.show_zero_values = rdsheet.show_zero_values
        wtsheet.auto_colour_grid = rdsheet.automatic_grid_line_colour
        wtsheet.cols_right_to_left = rdsheet.columns_from_right_to_left
        wtsheet.show_outline = rdsheet.show_outline_symbols
        wtsheet.remove_splits = rdsheet.remove_splits_if_pane_freeze_is_removed
        wtsheet.selected = rdsheet.sheet_selected
        # xlrd doesn't read WINDOW1 records, so we have to make a best
        # guess at which the active sheet should be:
        # (at a guess, only one sheet should be marked as visible)
        """
        if not self.sheet_visible and rdsheet.sheet_visible:
            self.wtbook.active_sheet = self.wtsheet_index
            wtsheet.sheet_visible = 1
        self.wtsheet_index += 1
        """

        wtsheet.page_preview = rdsheet.show_in_page_break_preview
        wtsheet.first_visible_row = rdsheet.first_visible_rowx
        wtsheet.first_visible_col = rdsheet.first_visible_colx
        wtsheet.grid_colour = rdsheet.gridline_colour_index
        wtsheet.preview_magn = rdsheet.cooked_page_break_preview_mag_factor
        wtsheet.normal_magn = rdsheet.cooked_normal_view_mag_factor
        #
        # DEFAULTROWHEIGHT
        #
        if rdsheet.default_row_height is not None:
            wtsheet.row_default_height = rdsheet.default_row_height
        wtsheet.row_default_height_mismatch = rdsheet.default_row_height_mismatch
        wtsheet.row_default_hidden = rdsheet.default_row_hidden
        wtsheet.row_default_space_above = rdsheet.default_additional_space_above
        wtsheet.row_default_space_below = rdsheet.default_additional_space_below
        #
        # BOUNDSHEET
        #
        wtsheet.visibility = rdsheet.visibility

        #
        # PANE
        #
        if rdsheet.has_pane_record:
            wtsheet.split_position_units_are_twips = True
            wtsheet.active_pane = rdsheet.split_active_pane
            wtsheet.horz_split_pos = rdsheet.horz_split_pos
            wtsheet.horz_split_first_visible = rdsheet.horz_split_first_visible
            wtsheet.vert_split_pos = rdsheet.vert_split_pos
            wtsheet.vert_split_first_visible = rdsheet.vert_split_first_visible

        # print settings
        if hasattr(rdsheet, 'print_headers'):
            wtsheet.print_headers = rdsheet.print_headers
            wtsheet.print_grid = rdsheet.print_grid
            wtsheet.vert_page_breaks = rdsheet.vertical_page_breaks
            wtsheet.horz_page_breaks = rdsheet.horizontal_page_breaks
            wtsheet.header_str = rdsheet.header_str
            wtsheet.footer_str = rdsheet.footer_str
            wtsheet.print_centered_vert = rdsheet.print_centered_vert
            wtsheet.print_centered_horz = rdsheet.print_centered_horz
            wtsheet.left_margin = rdsheet.left_margin
            wtsheet.right_margin = rdsheet.right_margin
            wtsheet.top_margin = rdsheet.top_margin
            wtsheet.bottom_margin = rdsheet.bottom_margin
            wtsheet.paper_size_code = rdsheet.paper_size_code
            wtsheet.print_scaling = rdsheet.print_scaling
            wtsheet.start_page_number = rdsheet.start_page_number
            wtsheet.fit_width_to_pages = rdsheet.fit_width_to_pages
            wtsheet.fit_height_to_pages = rdsheet.fit_height_to_pages
            wtsheet.print_in_rows = rdsheet.print_in_rows
            wtsheet.portrait = rdsheet.portrait
            wtsheet.print_colour = not rdsheet.print_not_colour
            wtsheet.print_draft = rdsheet.print_draft
            wtsheet.print_notes = rdsheet.print_notes
            wtsheet.print_notes_at_end = rdsheet.print_notes_at_end
            wtsheet.print_omit_errors = rdsheet.print_omit_errors
            wtsheet.print_hres = rdsheet.print_hres
            wtsheet.print_vres = rdsheet.print_vres
            wtsheet.header_margin = rdsheet.header_margin
            wtsheet.footer_margin = rdsheet.footer_margin
            wtsheet.copies_num = rdsheet.copies_num

    def copy_row_dimension(self, rdrowx, wtrowx):
        """
        This should be called every time processing of a new
        row in the current sheet starts.

        :param rdrowx: the index of the row in the current sheet from which
                 information for the row to be written will be
                 copied.

        :param wtrowx: the index of the row in sheet to be written to which
                 information will be written for the row being read.
        """
        if wtrowx in self.wtrows or rdrowx not in self.rdsheet.rowinfo_map:
            return
        wtrow = self.wtsheet.row(wtrowx)
        # empty rows may not have a rowinfo record
        rdrow = self.rdsheet.rowinfo_map.get(rdrowx)
        if rdrow:
            wtrow.height = rdrow.height
            wtrow.has_default_height = rdrow.has_default_height
            wtrow.height_mismatch = rdrow.height_mismatch
            wtrow.level = rdrow.outline_level
            wtrow.collapse = rdrow.outline_group_starts_ends  # No kiddin'
            wtrow.hidden = rdrow.hidden
            wtrow.space_above = rdrow.additional_space_above
            wtrow.space_below = rdrow.additional_space_below
            if rdrow.has_default_xf_index:
                wtrow.set_style(self.style_list[rdrow.xf_index])
            self.wtrows.add(wtrowx)

    def copy_col_dimension(self, rdcolx, wtcolx):
        if wtcolx not in self.wtcols and rdcolx in self.rdsheet.colinfo_map:
            rdcol = self.rdsheet.colinfo_map[rdcolx]
            wtcol = self.wtsheet.col(wtcolx)
            wtcol.width = rdcol.width
            wtcol.set_style(self.style_list[rdcol.xf_index])
            wtcol.hidden = rdcol.hidden
            wtcol.level = rdcol.outline_level
            wtcol.collapsed = rdcol.collapsed
            self.wtcols.add(wtcolx)

    def _cell(self, source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value=None, cty=None):
        if value is None:
            value = source_cell.value
            cty = source_cell.ctype
        if cty is None:
            value,cty = get_type(value)

        if cty == xlrd.XL_CELL_EMPTY:
            return
        if source_cell.xf_index is not None:
            style = self.style_list[source_cell.xf_index]
        else:
            style = self.style_list[0]

        wtrow = self.wtsheet.row(wtrowx)
        if cty == xlrd.XL_CELL_TEXT:
            if isinstance(value, (list, tuple)):
                wtrow.set_cell_rich_text(wtcolx, value, style)
            else:
                wtrow.set_cell_text(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_NUMBER or cty == xlrd.XL_CELL_DATE:
            wtrow.set_cell_number(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_BLANK:
            wtrow.set_cell_blank(wtcolx, style)
        elif cty == xlrd.XL_CELL_BOOLEAN:
            wtrow.set_cell_boolean(wtcolx, value, style)
        elif cty == xlrd.XL_CELL_ERROR:
            wtrow.set_cell_error(wtcolx, value, style)
        else:
            raise Exception(
                "Unknown xlrd cell type %r with value %r at (sheet=%r,rowx=%r,colx=%r)" \
                % (cty, value, self.rdsheet.name, rdrowx, rdcolx)
            )

    def cell(self, source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value=None, cty=None):
        self.copy_row_dimension(rdrowx, wtrowx)
        self.copy_col_dimension(rdcolx, wtcolx)
        self._cell(source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value, cty)


class BookBase():

    def load_rdbook(self, fname):
        self.rdbook = rdbook = xlrd.open_workbook(fname, formatting_info=True)
        self.style_list = []
        self.font_map = {}
        for rdxf in rdbook.xf_list:
            wtxf = xlwt.Style.XFStyle()
            #
            # number format
            #
            wtxf.num_format_str = rdbook.format_map[rdxf.format_key].format_str
            #
            # font
            #
            wtf = wtxf.font
            rdf = rdbook.font_list[rdxf.font_index]
            wtf.height = rdf.height
            wtf.italic = rdf.italic
            wtf.struck_out = rdf.struck_out
            wtf.outline = rdf.outline
            wtf.shadow = rdf.outline
            wtf.colour_index = rdf.colour_index
            wtf.bold = rdf.bold  #### This attribute is redundant, should be driven by weight
            wtf._weight = rdf.weight  #### Why "private"?
            wtf.escapement = rdf.escapement
            wtf.underline = rdf.underline_type  ####
            # wtf.???? = rdf.underline #### redundant attribute, set on the fly when writing
            wtf.family = rdf.family
            wtf.charset = rdf.character_set
            wtf.name = rdf.name
            self.font_map[rdxf.font_index] = wtf
            #
            # protection
            #
            wtp = wtxf.protection
            rdp = rdxf.protection
            wtp.cell_locked = rdp.cell_locked
            wtp.formula_hidden = rdp.formula_hidden
            #
            # border(s) (rename ????)
            #
            wtb = wtxf.borders
            rdb = rdxf.border
            wtb.left = rdb.left_line_style
            wtb.right = rdb.right_line_style
            wtb.top = rdb.top_line_style
            wtb.bottom = rdb.bottom_line_style
            wtb.diag = rdb.diag_line_style
            wtb.left_colour = rdb.left_colour_index
            wtb.right_colour = rdb.right_colour_index
            wtb.top_colour = rdb.top_colour_index
            wtb.bottom_colour = rdb.bottom_colour_index
            wtb.diag_colour = rdb.diag_colour_index
            wtb.need_diag1 = rdb.diag_down
            wtb.need_diag2 = rdb.diag_up
            #
            # background / pattern (rename???)
            #
            wtpat = wtxf.pattern
            rdbg = rdxf.background
            wtpat.pattern = rdbg.fill_pattern
            wtpat.pattern_fore_colour = rdbg.pattern_colour_index
            wtpat.pattern_back_colour = rdbg.background_colour_index
            #
            # alignment
            #
            wta = wtxf.alignment
            rda = rdxf.alignment
            wta.horz = rda.hor_align
            wta.vert = rda.vert_align
            wta.dire = rda.text_direction
            # wta.orie # orientation doesn't occur in BIFF8! Superceded by rotation ("rota").
            wta.rota = rda.rotation
            wta.wrap = rda.text_wrapped
            wta.shri = rda.shrink_to_fit
            wta.inde = rda.indent_level
            # wta.merg = ????
            #
            self.style_list.append(wtxf)


    def create_workbook(self):
        self.wtbook = xlwt.Workbook(style_compression=2)
        self.wtbook.dates_1904 = self.rdbook.datemode
        self.wtbook.wtsheet_names = set()

        # Set the default style and the default font
        idx = self.wtbook.add_style(None)
        if idx == 0x10:
            return
        for idx, wtxf in enumerate(self.style_list):
            self.wtbook.add_style(wtxf)
            if idx == 15:
                return
        wtxf = self.style_list[0]
        for _ in range(15 - idx):
            self.wtbook.add_style(wtxf)
        return self.wtbook

    def _get_font(self, index):
        wtf = self.font_map.get(index)
        if wtf:
            return wtf
        else:
            wtf = xlwt.Font()
            rdf = self.rdbook.font_list[index]
            wtf.height = rdf.height
            wtf.italic = rdf.italic
            wtf.struck_out = rdf.struck_out
            wtf.outline = rdf.outline
            wtf.shadow = rdf.outline
            wtf.colour_index = rdf.colour_index
            wtf.bold = rdf.bold
            wtf._weight = rdf.weight
            wtf.escapement = rdf.escapement
            wtf.underline = rdf.underline_type
            wtf.family = rdf.family
            wtf.charset = rdf.character_set
            wtf.name = rdf.name
            self.font_map[index] = wtf
            return wtf

    def get_font(self, sheet, rowx, colx):
        xf = self.rdbook.xf_list[sheet.cell_xf_index(rowx, colx)]
        return self._get_font(xf.font_index)

    def get_rich_text(self, sheet, rowx, colx):
        cell_value = sheet.cell_value(rowx, colx)
        if not cell_value:
            return
        runlist = sheet.rich_text_runlist_map.get((rowx, colx))
        if runlist:
            rich_text = []
            for idx,(start,font_idx) in enumerate(runlist):
                end = None
                if idx != len(runlist) - 1:
                    end = runlist[idx + 1][0]
                text = cell_value[start:end]
                font = self._get_font(font_idx)
                rich_text.append((text, font))
            if runlist[0][0] != 0:
                text = cell_value[:runlist[0][0]]
                xf = self.rdbook.xf_list[sheet.cell_xf_index(rowx, colx)]
                font = self._get_font(xf.font_index)
                rich_text.insert(0, (text, font))
            return rich_text




class SheetMixin(object):

    def reset_pos(self):
        self.current_row_num = self.min_rowx
        self.current_col_num = self.min_colx

    def set_sheet_resource(self, sheet_resource):
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet

    def write_row(self, row_node):
        self.current_row_num += 1
        self.current_col_num = self.min_colx

    def write_cell(self, cell_node, rv, cty):
        self.current_col_num += 1
        self.merger.merge_cell(cell_node.rowx, cell_node.colx, self.current_row_num, self.current_col_num)
        if cell_node.sheet_cell:
            self.cell(cell_node.sheet_cell, cell_node.rowx, cell_node.colx, self.current_row_num, self.current_col_num, rv, cty)

    def set_image_ref(self, image_ref, image_key):
        wt_top_left = (self.current_row_num, self.current_col_num+1)
        self.merger.set_image_ref(image_ref, image_key, wt_top_left)


class BookMixin(object):

    def load(self, fname):
        pass

    def build(self, sheet, index):
        pass

    def set_jinja_globals(self, **kwargs):
        self.jinja_env.globals.update(kwargs)

    def get_sheet_writer(self, sheet_resource, sheet_name):
        sheet_writer = self.sheet_writer_map.get(sheet_name)
        if not sheet_writer:
            sheet_writer = self.sheet_writer_cls(self, sheet_resource, sheet_name)
            self.sheet_writer_map[sheet_name] = sheet_writer
        else:
            sheet_writer.set_sheet_resource(sheet_resource)
        return sheet_writer

    def get_sheet_name(self, payload):
        sheet_name = payload.get('sheet_name')
        if sheet_name:
            return sheet_name
        for i in range(9999):
            sheet_name = "sheet%d" % i
            if not self.sheet_writer_map.get(sheet_name):
                return sheet_name
        return "XLSheet"

    def put_sheet_resource(self, index, name, sheet_resource):
        self.sheet_resource_map[index] = sheet_resource
        self.sheet_resource_map[name] = sheet_resource

    def get_sheet_resource(self, payload):
        key = payload.get('tpl_name') or payload.get('tpl_idx') or payload.get('tpl_index')
        return self.sheet_resource_map.get(key) or self.sheet_resource_map.get(0)

    def render_sheet(self, payload):
        sheet_name = self.get_sheet_name(payload)
        sheet_resource = self.get_sheet_resource(payload)
        sheet_writer = self.get_sheet_writer(sheet_resource, sheet_name)
        sheet_resource.render_sheet(sheet_writer, payload)

    def render_sheets(self, payloads):
        for payload in payloads:
            self.render_sheet(payload)

    def render_book(self, payloads):
        return self.render_sheets(payloads)

    def save(self, fname):
        pass


# -*- coding: utf-8 -*-

import copy
from openpyxl.utils import get_column_letter
from openpyxl.cell.text import InlineFont

from openpyxl.cell.cell import NUMERIC_TYPES, TIME_TYPES, STRING_TYPES
BOOL_TYPE = bool

def get_type(value):
    if isinstance(value, NUMERIC_TYPES):
        dt = 'n'
    elif isinstance(value, STRING_TYPES):
        dt = 's'
    elif isinstance(value, TIME_TYPES):
        dt = 'd'
    elif isinstance(value, BOOL_TYPE):
        dt = 'b'
    else:
        return str(value), 's'
    return value, dt

class SheetBase():

    def copy_sheet_settings(self):
        self.wtsheet.sheet_format = copy.copy(self.rdsheet.sheet_format)
        self.wtsheet.sheet_properties = copy.copy(self.rdsheet.sheet_properties)
        # copy print settings
        self.wtsheet.page_setup = copy.copy(self.rdsheet.page_setup)
        self.wtsheet.print_options = copy.copy(self.rdsheet.print_options)
        self.wtsheet._print_rows = copy.copy(self.rdsheet._print_rows)
        self.wtsheet._print_cols = copy.copy(self.rdsheet._print_cols)
        self.wtsheet._print_area = copy.copy(self.rdsheet._print_area)
        self.wtsheet.page_margins = copy.copy(self.rdsheet.page_margins)
        self.wtsheet.protection = copy.copy(self.rdsheet.protection)
        self.wtsheet.HeaderFooter = copy.copy(self.rdsheet.HeaderFooter)
        self.wtsheet.views = copy.copy(self.rdsheet.views)
        self.wtsheet._images = copy.copy(self.rdsheet._images)

    def copy_row_dimension(self, rdrowx, wtrowx):
        if wtrowx in self.wtrows:
            return
        dim = self.rdsheet.row_dimensions.get(rdrowx)
        if dim:
            self.wtsheet.row_dimensions[wtrowx] = copy.copy(dim)
            self.wtsheet.row_dimensions[wtrowx].worksheet = self.wtsheet
            self.wtrows.add(wtrowx)

    def copy_col_dimension(self, rdcolx, wtcolx):
        if wtcolx in self.wtcols:
            return
        rdkey = get_column_letter(rdcolx)
        rddim = self.rdsheet.column_dimensions.get(rdkey)
        if not rddim:
            return
        wtdim = copy.copy(rddim)
        if rdcolx != wtcolx:
            wtkey = get_column_letter(wtcolx)
            wtdim.index = wtkey
            d = wtcolx - rdcolx
            wtdim.min += d
            wtdim.max += d
        else:
            wtkey = rdkey
        self.wtsheet.column_dimensions[wtkey] = wtdim
        self.wtsheet.column_dimensions[wtkey].worksheet = self.wtsheet
        self.wtcols.add(wtcolx)

    def _cell(self, source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value=None, data_type=None):
        target_cell = self.wtsheet.cell(column=wtcolx, row=wtrowx)
        if value is None:
            target_cell.value = source_cell._value
            target_cell.data_type = source_cell.data_type
        elif data_type:
            target_cell._value = value
            target_cell.data_type = data_type
        else:
            value, data_type = get_type(value)
            target_cell.value = value
            target_cell.data_type = data_type
        if source_cell.has_style:
            target_cell._style = copy.copy(source_cell._style)
        if source_cell.hyperlink:
            target_cell._hyperlink = copy.copy(source_cell.hyperlink)
        #if source_cell.comment:
        #    target_cell.comment = copy.copy(source_cell.comment)

    def cell(self, source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value=None, data_type=None):
        self.copy_row_dimension(rdrowx, wtrowx)
        self.copy_col_dimension(rdcolx, wtcolx)
        self._cell(source_cell, rdrowx, rdcolx, wtrowx, wtcolx, value, data_type)

class BookBase():

    def get_font(self, fontId):
        ifont = self.font_map.get(fontId)
        if ifont:
            return ifont
        else:
            font = self.workbook._fonts[fontId]
            ifont = InlineFont()
            ifont.rFont = font.name
            ifont.charset = font.charset
            ifont.family = font.family
            ifont.b = font.b
            ifont.i = font.i
            ifont.strike = font.strike
            ifont.outline = font.outline
            ifont.shadow = font.shadow
            ifont.condense = font.condense
            ifont.extend = font.extend
            ifont.color = font.color
            ifont.sz = font.sz
            ifont.u = font.u
            ifont.vertAlign = font.vertAlign
            ifont.scheme = font.scheme
            self.font_map[fontId] = ifont
            return ifont



# -*- coding: utf-8 -*-

from .patchx import *
from openpyxl import load_workbook
from .basex import SheetBase, BookBase
from .writermixin import SheetMixin, BookMixin
from .utils import tag_test
from .xlnode import Tree, Row, Cell, EmptyCell, Node, create_cell
from .jinja import jinja_envx
from .sheetresource import SheetResource
from .richtexthandler import rich_handlerx
from .mergerx import Merger

class SheetWriter(SheetBase, SheetMixin):

    def __init__(self, bookwriter, sheet_resource, sheet_name):
        self.workbook = bookwriter.workbook
        self.merger = sheet_resource.merger
        self.rdsheet = sheet_resource.rdsheet
        self.wtsheet = self.workbook.create_sheet(title=sheet_name)
        self.copy_sheet_settings()
        self.wtrows = set()
        self.wtcols = set()
        self.min_rowx = 0
        self.min_colx = 0
        self.reset_pos()


class BookWriter(BookBase, BookMixin):
    sheet_writer_cls = SheetWriter

    def __init__(self, fname, debug=False):
        self.debug = debug
        self.load(fname)


    def load(self, fname):
        self.workbook = load_workbook(fname)
        self.font_map = {}
        self.sheet_writer_map = {}
        self.sheet_resource_map = {}
        self.jinja_env = jinja_envx
        for index,rdsheet in enumerate(self.workbook.worksheets):
            sheet_tree = self.build(rdsheet, index)
            merger = Merger(rdsheet)
            sheet_resource = SheetResource(rdsheet, sheet_tree, jinja_envx, merger)
            self.put_sheet_resource(index, rdsheet.title, sheet_resource)
            self.workbook.remove(rdsheet)

    def build(self, sheet, index):
        tree = Tree(index)

        for rowx in range(1, sheet.max_row + 1):
            row_node = Row(rowx)
            tree.add_child(row_node)
            for colx in range(1, sheet.max_column + 1):
                sheet_cell = sheet._cells.get((rowx, colx))
                if not sheet_cell:
                    cell_node = EmptyCell(rowx, colx)
                    tree.add_child(cell_node)
                    continue

                value = sheet_cell._value
                data_type = sheet_cell.data_type
                rich_text = None
                if hasattr(value, 'rich') and value.rich:
                    rich_text = value.rich
                if data_type == 's':
                    if not tag_test(value):
                        cell_node = Cell(sheet_cell, rowx, colx, value, data_type)
                    else:
                        font = self.get_font(sheet_cell._style.fontId)
                        cell_node = create_cell(sheet_cell, rowx, colx, value, rich_text, data_type, font, rich_handlerx)
                else:
                    cell_node = Cell(sheet_cell, rowx, colx, value, data_type)
                tree.add_child(cell_node)
            tree.add_child(Node())#
        return tree


    def save(self, fname):
        if not self.workbook.active:
            self.workbook.active = 0
        self.workbook.save(fname)
        for sheet in self.workbook.worksheets:
            self.workbook.remove(sheet)
        self.sheet_writer_map.clear()



# -*- coding: utf-8 -*-

import six
from openpyxl.cell.text import Text

"""
A temporary workaround to support rich text.
"""
class IAMRich(six.text_type):

    def __new__(cls, value, rich):
        obj = six.text_type.__new__(cls, value)
        obj.rich = rich
        return obj

    def replace(self, *args, **kwargs):
        s = super(self.__class__, self).replace(*args, **kwargs)
        if s == self:
            return self
        return IAMRich(s, self.rich)

class RichText(Text):

    @property
    def content(self):
        """
        Text stripped of all formatting
        """
        snippets = []
        if self.plain is not None:
            snippets.append(self.plain)
        isrich = False
        for block in self.formatted:
            if block.t is not None:
                isrich = True
                snippets.append(block.t)
        rv = u"".join(snippets)
        if isrich:
            return IAMRich(rv, self)
        return rv

from openpyxl.compat import safe_string
from openpyxl.xml.functions import Element, SubElement, whitespace, XML_NS
from openpyxl import LXML
from openpyxl.cell._writer import _set_attributes

def etree_write_cell(xf, worksheet, cell, styled=None):

    value, attributes = _set_attributes(cell, styled)

    el = Element("c", attributes)
    if value is None or value == "":
        xf.write(el)
        return

    if cell.data_type == 'f':
        shared_formula = worksheet.formula_attributes.get(cell.coordinate, {})
        formula = SubElement(el, 'f', shared_formula)
        if value is not None:
            formula.text = value[1:]
            value = None

    if cell.data_type == 's':
        if hasattr(value, 'rich') and value.rich:
            sub = value.rich.to_tree(tagname='is')
            el.append(sub)
        else:
            inline_string = SubElement(el, 'is')
            text = SubElement(inline_string, 't')
            text.text = value
            whitespace(text)

    else:
        cell_content = SubElement(el, 'v')
        if value is not None:
            cell_content.text = safe_string(value)

    xf.write(el)


def lxml_write_cell(xf, worksheet, cell, styled=False):
    value, attributes = _set_attributes(cell, styled)

    if value == '' or value is None:
        with xf.element("c", attributes):
            return

    with xf.element('c', attributes):
        if cell.data_type == 'f':
            shared_formula = worksheet.formula_attributes.get(cell.coordinate, {})
            with xf.element('f', shared_formula):
                if value is not None:
                    xf.write(value[1:])
                    value = None

        if cell.data_type == 's':
            if hasattr(value, 'rich') and value.rich:
                el = value.rich.to_tree(tagname='is')
                xf.write(el)
            else:
                with xf.element("is"):
                    attrs = {}
                    if value != value.strip():
                        attrs["{%s}space" % XML_NS] = "preserve"
                    el = Element("t", attrs)  # lxml can't handle xml-ns
                    el.text = value
                    xf.write(el)
                    # with xf.element("t", attrs):
                    # xf.write(value)
        else:
            with xf.element("v"):
                if value is not None:
                    xf.write(safe_string(value))

Text.content = RichText.content
if LXML:
    write_cell = lxml_write_cell
else:
    write_cell = etree_write_cell

from openpyxl.worksheet import _writer
_writer.write_cell = write_cell

"""
fix  image error
"""
from openpyxl.xml.constants import DRAWING_NS
from openpyxl.drawing.geometry import GeomGuideList
GeomGuideList.tagname = "avLst"
GeomGuideList.namespace = DRAWING_NS





class NodeMap():
    node_map = {}

    def __init__(self):
        self.current_node = None
        self.current_key = ''
        self.last_node = None
        self.last_key = None

    def set_current_node(self, node):
        self.current_node = node
        self.current_key = node.node_key

    def put(self, key, node):
        self.node_map[key] = node

    def find_lca(self, pre, next):
        # find lowest common ancestor
        next_branch = []
        if pre.depth > next.depth:
            for i in range(next.depth, pre.depth):
                pre.exit()
                # print(pre, 'pre up', pre._parent)
                pre = pre._parent

        elif pre.depth < next.depth:
            for i in range(pre.depth, next.depth):
                next_branch.insert(0, next)
                # print(next, 'next up', next._parent)
                next = next._parent
        if pre is next:
            pass
        else:
            pre_parent = pre._parent
            next_parent = next._parent
            while pre_parent != next_parent:
                # print(pre, next, 'up together')
                pre.exit()
                pre = pre_parent
                pre_parent = pre._parent
                next_branch.insert(0, next)
                next = next_parent
                next_parent = next._parent
            pre.exit()
            if pre_parent._children.index(pre) > pre_parent._children.index(next):
                pre_parent.child_reenter()
                next.reenter()
            else:
                next.enter()

        for next in next_branch:
            next.enter()


    def get_node(self, key):
        if key == self.current_key:
            return self.current_node
        else:
            self.last_key = self.current_key
            self.last_node = self.current_node
            self.current_node = self.node_map.get(key)
            self.current_key = key
            self.find_lca(self.last_node, self.current_node)
        return self.current_node

node_map = NodeMap()